# -*- coding: utf-8 -*-

from __future__ import absolute_import, print_function

import argparse
import itertools
import json
import logging
import os
import six

from pprint import pprint

from rhmsg.activemq.consumer import AMQConsumer
from rhmsg.cli.common import add_common_cli_args
from .envs import broker_envs

logging.basicConfig(level=logging.CRITICAL,
                    format='[%(levelname)s] %(message)s')
logger = logging.getLogger(__name__)


def build_command_line_parser():
    parser = argparse.ArgumentParser(
        description='Command line tool to consume messages from remote broker.'
    )
    add_common_cli_args(parser)
    parser.add_argument(
        '--address',
        dest='address',
        help='Message will be received from this address. An address is in '
             'format `Consumer.some-name.VirtualTopic.topic`. In practice, '
             'please consult the infra team to determine what value should '
             'substitute the part some-name. For technical detailed '
             'information, refer to documentation Virtual Topics of ActiveMQ.')
    parser.add_argument(
        '--durable',
        action='store_true',
        default=False,
        help='Create a durable subscription')
    parser.add_argument(
        '--subscription-name',
        help='Use the specified name for the subscription, otherwise it will '
             'be generated. Implies --durable.')
    parser.add_argument(
        '--dump',
        action='store_true',
        default=False,
        dest='dump_message',
        help='Display more detailed information except the message body sent '
             'by the original sender.')
    parser.add_argument(
        '--selector',
        dest='selector',
        default=None,
        help='A formatted string to filter messages based on headers. Refer to'
             ' documentation for JMS Selectors. A quick example is '
             '"type = \'StateChange\'".')
    parser.add_argument(
        '--pp',
        dest='pretty_print',
        action='store_true',
        default=False,
        help='Pretty-print received messages.')
    parser.add_argument(
        '--manual-ack',
        action='store_true',
        default=False,
        help='Accept and ack received message manually.')
    parser.add_argument(
        '--one-message-only',
        action='store_true',
        default=False,
        help='Only consume one message and exit.')

    return parser


counter = itertools.count(1)


def message_handler(message, data):
    num = six.next(counter)

    body = message.body
    if isinstance(body, six.text_type):
        body = body.encode('utf-8', 'backslashreplace')
    if data['dump']:
        print('------------- ({0}) {1} --------------'.format(num, message.id))
        print('address:', message.address)
        print('subject:', message.subject)
        print('properties:', message.properties)
        print('durable:', message.durable)
        print('content_type:', message.content_type)
        print('content_encoding:', message.content_encoding)
        print('delivery_count:', message.delivery_count)
        print('reply_to:', message.reply_to)
        print('priority:', message.priority)
        if data['pp']:
            print('body:')
            pprint(json.loads(body))
        else:
            print('body:', body)
    else:
        if data['pp']:
            print('Got [%02d]:' % num)
            pprint(json.loads(body))
        else:
            print('Got [%02d]:' % num, body)

    return data['one_message_only'], not data['manual_ack']


def consume_messages(args):
    consumer_config = {
        'urls': broker_envs[args.env],
        'certificate': args.cert_file,
        'private_key': args.private_key_file,
        'trusted_certificates': args.ca_certs,
        }
    if args.durable and not args.subscription_name:
        args.subscription_name = args.address
    consumer = AMQConsumer(**consumer_config)
    consumer.consume(args.address,
                     selector=args.selector,
                     callback=message_handler,
                     auto_accept=False,
                     subscription_name=args.subscription_name,
                     data={'dump': args.dump_message,
                           'pp': args.pretty_print,
                           'one_message_only': args.one_message_only,
                           'manual_ack': args.manual_ack,
                           })


def main():
    try:
        parser = build_command_line_parser()
        args = parser.parse_args()

        if args.debug:
            os.environ['PN_TRACE_FRM'] = '1'

        consume_messages(args)
    except KeyboardInterrupt:
        pass
