# -*- coding: utf-8 -*-

from __future__ import absolute_import

import argparse
import logging
import proton
import os
import six

from rhmsg.activemq.producer import AMQProducer
from rhmsg.cli.common import add_common_cli_args
from .envs import broker_envs

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)


def build_command_line_parser():
    parser = argparse.ArgumentParser(
        description='Command line tool to send message via remote broker.'
    )
    add_common_cli_args(parser)
    parser.add_argument(
        '--queue',
        dest='producer_queue',
        help='Queue to send messages.')
    parser.add_argument(
        '--topic',
        dest='producer_topic',
        help='Topic to send messages.')
    parser.add_argument(
        '--subject',
        dest='subject',
        default='amqproducer: hello',
        help="Message's subject")
    parser.add_argument(
        '--property',
        action='append',
        dest='properties',
        help='Specify particular properties for the outgoing message. Each '
             'property is a pair of name and value in form name=value. Can be '
             'specified multiple times to set more than one property for an '
             'outgoing message.')
    parser.add_argument(
        'message_body',
        nargs='+',
        help='The message body to be sent.')

    return parser


def produce_message(args):
    producer_config = {
        'urls': broker_envs[args.env],
        'certificate': args.cert_file,
        'private_key': args.private_key_file,
        'trusted_certificates': args.ca_certs,
        }

    if args.properties:
        properties = dict((prop.split('=', 1) for prop in args.properties))
    else:
        properties = {}

    with AMQProducer(**producer_config) as producer:
        if args.producer_queue:
            producer.through_queue(args.producer_queue)
        if args.producer_topic:
            producer.through_topic(args.producer_topic)

        message = proton.Message()
        message.body = ' '.join(args.message_body)
        if isinstance(message.body, six.binary_type):
            # On python2, convert to unicode
            message.body = message.body.decode('utf-8', 'replace')
        message.properties = properties.copy()
        message.subject = args.subject
        producer.send(message)


def main():
    try:
        parser = build_command_line_parser()
        args = parser.parse_args()

        if args.debug:
            os.environ['PN_TRACE_FRM'] = '1'

        produce_message(args)
    except KeyboardInterrupt:
        pass
