# -*- coding: utf-8 -*-

"""
Get UMB messages from internal datagrepper and send it to given broker.
"""

from __future__ import absolute_import, print_function

import argparse
import json
import os
import logging
import requests
import proton
import sys

from rhmsg.activemq.producer import AMQProducer
from rhmsg.cli.common import add_common_cli_args
from .envs import broker_envs

logging.basicConfig(level=logging.CRITICAL,
                    format='[%(levelname)s] %(message)s')
logger = logging.getLogger(__name__)


class InvalidMessageError(Exception):
    def __init__(self, original_error, line_no):
        self.original_error = original_error
        self.line_no = line_no

    def __str__(self):
        return 'Cannot decode message from file. {0}'.format(
            str(self.original_error))

    def __repr__(self):
        return '<{0}: at line {1}>'.format(
            self.__class__.__name__, self.line_no)


def validate_arg_msg_file(value):
    if not os.path.exists(value):
        raise argparse.ArgumentTypeError(
            'Message file {0} does not exist.'.format(value))
    return value


def build_command_line_parser():
    parser = argparse.ArgumentParser(
        description='Read unified messages from datagrepper or a file '
                    'replay to unified message bus. If no file is specified, '
                    'messages will be retrieved from datagrepper.'
    )
    add_common_cli_args(parser)
    parser.add_argument(
        '--address',
        dest='address',
        required=True,
        help='An address of a queue or topic the message will be resent to. '
             'For detailed information of topic, please refer to ActiveMQ '
             'documentation "Virtual Destinations".')

    exclusive_group = parser.add_mutually_exclusive_group(required=True)
    exclusive_group.add_argument(
        '--server-url',
        dest='server_url',
        help='Server URL of datagrepper. Example: http://hostname[:port]/')
    exclusive_group.add_argument(
        '-f', '--file',
        metavar='FILE',
        dest='msg_file',
        type=validate_arg_msg_file,
        help='File containing message to replay. Multiple messages can be '
             'listed line by line. Each message must be in valid JSON format.')

    parser.add_argument(
        '--no-ssl-verify',
        dest='no_ssl_verify',
        action='store_true',
        default=False,
        help='Do not verify SSL when connecting to remote datagrepper instance.')

    parser.add_argument(
        'msg_id',
        nargs='*',
        help='The message ID. This value could be retrieved from a ActiveMQ'
             ' raw message body by key msg_id. This is optional if --file is '
             'specified that means to read message from a specific file. '
             'Multiple message IDs can be specified.'
    )

    return parser


def grep_messages(server_url, msg_id, verify_ssl=True):
    url = server_url.strip('/')
    params = {
        'id': msg_id,
        'is_raw': 'true',
        'size': 'extra-large',
    }
    response = requests.get(url, params=params, verify=verify_ssl)
    if response.status_code != 200:
        raise RuntimeError(response.content)
    return response.json()


def load_messages_from_file(filename):
    with open(filename, 'r') as fin:
        for line_no, line in enumerate(fin, 1):
            try:
                data = json.loads(line)
            # json.loads raises different errors in Python 2 and 3. So, here
            # just catch whatever it raises for easy compatibility.
            except Exception as e:  # noqa
                raise InvalidMessageError(e, line_no)
            else:
                yield data


def replay_messages(msg, broker_urls, topic, cert_file, private_key_file, ca_certs):
    producer_config = {
        'urls': broker_urls,
        'certificate': cert_file,
        'private_key': private_key_file,
        'trusted_certificates': ca_certs,
        }

    with AMQProducer(**producer_config) as producer:
        producer.through_topic(topic)

        message = proton.Message()
        message.subject = 'Replay {0}'.format(msg['msg_id'])
        message.body = json.dumps(msg)

        producer.send(message)


def main():
    parser = build_command_line_parser()
    args = parser.parse_args()

    if args.msg_file and args.msg_id:
        print('Ambiguous arguments: both --file and message IDs are given. '
              'Please choose one.',
              file=sys.stderr)
        sys.exit(1)

    if args.server_url and not args.msg_id:
        print('Missing message ID.', file=sys.stderr)
        print('When specifying --server-url to retrieve message from remote '
              'datagrepper, message ID must be specified.',
              file=sys.stderr)
        sys.exit(1)

    if args.debug:
        os.environ['PN_TRACE_FRM'] = '1'

    if args.msg_file:
        msgs = load_messages_from_file(args.msg_file)
    else:
        msgs = [
            grep_messages(args.server_url,
                          msg_id,
                          verify_ssl=not args.no_ssl_verify)
            for msg_id in args.msg_id
        ]

    for msg in msgs:
        replay_messages(msg=msg,
                        broker_urls=broker_envs[args.env],
                        topic=args.address,
                        cert_file=args.cert_file,
                        private_key_file=args.private_key_file,
                        ca_certs=args.ca_certs)
